<?php

/**
 * A PSR-3 inspired logger class
 * This class logs + formats logs for display in the Simple History GUI/Viewer
 *
 * Extend this class to make your own logger
 *
 * @link https://github.com/php-fig/fig-standards/blob/master/accepted/PSR-3-logger-interface.md PSR-3 specification
 */
class SimpleLogger
{

	/**
	 * Unique slug for this logger
	 * Will be saved in DB and used to associate each log row with its logger
	 */
	public $slug = __CLASS__;

	/**
	 * Will contain the untranslated messages from getInfo()
	 *
	 * By adding your messages here they will be stored both translated and non-translated
	 * You then log something like this:
	 * <code>
	 *   $this->info( $this->messages["POST_UPDATED"] );
	 * </code>
	 * or with the shortcut
	 * <code>
	 *   $this->infoMessage("POST_UPDATED");
	 * </code>
	 * which results in the original, untranslated, string being added to the log and database
	 * the translated string are then only used when showing the log in the GUI
	 */
	public $messages;

	/**
	 * ID of last inserted row. Used when chaining methods.
	 */
	private $lastInsertID;

	/**
	 * Constructor. Remember to call this as parent constructor if making a childlogger
	 * @param $simpleHistory history class  objectinstance
	 */
	public function __construct($simpleHistory) {

		global $wpdb;

		$this->db_table = $wpdb->prefix . SimpleHistory::DBTABLE;
		$this->db_table_contexts = $wpdb->prefix . SimpleHistory::DBTABLE_CONTEXTS;

		$this->simpleHistory = $simpleHistory;

	}

	/**
	 * Method that is called automagically when logger is loaded by Simple History
	 * Add your init stuff here
	 */
	public function loaded() {

	}

	/**
	 * Get array with information about this logger
	 *
	 * @return array
	 */
	function getInfo() {

		$arr_info = array(

			// The logger slug. Defaulting to the class name is nice and logical I think
			"slug" => __CLASS__,

			// Shown on the info-tab in settings, use these fields to tell
			// an admin what your logger is used for
			"name" => "SimpleLogger",
			"description" => "The built in logger for Simple History",

			// Capability required to view log entries from this logger
			"capability" => "edit_pages",
			"messages" => array(
				// No pre-defined variants
				// when adding messages __() or _x() must be used
			)

		);

		return $arr_info;

	}

	/**
	 * Returns the capability required to read log rows from this logger
	 *
	 * @return $string capability
	 */
	public function getCapability() {

		$arr_info = $this->getInfo();

		return $arr_info["capability"];

	}

	/**
	* Interpolates context values into the message placeholders.
	*/
	function interpolate($message, $context = array())
	{

		if ( ! is_array($context) ) {
			return $message;
		}

		// build a replacement array with braces around the context keys
		$replace = array();
		foreach ($context as $key => $val) {
			$replace['{' . $key . '}'] = $val;
		}

		// interpolate replacement values into the message and return
		return strtr($message, $replace);

	}

	/**
	 * Returns header output for a log row
	 * Format should be common for all log rows and should be like:
	 * Username (user role) · Date
	 * @return string HTML
	 */
	function getLogRowHeaderOutput($row) {

		// HTML for initiator
		$initiator_html = "";

		$initiator = $row->initiator;
		$context = $row->context;

		switch ( $initiator ) {

			case "wp":
				$initiator_html .= '<strong class="SimpleHistoryLogitem__inlineDivided">WordPress</strong> ';
				break;

			// wp_user = wordpress uses, but user may have been deleted since log entry was added
			case "wp_user":

				$user_id = isset( $row->context["_user_id"] ) ? $row->context["_user_id"] : null;

				if ( $user_id > 0 && $user = get_user_by("id", $user_id) ) {

					// Sender is user and user still exists
					$is_current_user = ( $user_id == get_current_user_id() ) ? true : false;

					// get user role, as done in user-edit.php
					$user_roles = array_intersect( array_values( $user->roles ), array_keys( get_editable_roles() ) );
					$user_role  = array_shift( $user_roles );
					$user_display_name = $user->display_name;

					$tmpl_initiator_html = '
						<strong class="SimpleHistoryLogitem__inlineDivided">%3$s</strong>
						<span class="SimpleHistoryLogitem__inlineDivided SimpleHistoryLogitem__headerEmail">%2$s</span>
					';

					// If user who logged this is the currently logged in user
					// we replace name and email with just "You"
					if ($is_current_user) {
						$tmpl_initiator_html = '
							<strong class="SimpleHistoryLogitem__inlineDivided">%5$s</strong>
						';
					}

					/**
				     * Filter the format for the user output
				     *
				     * @since 2.0
				     *
				     * @param string $format.
				     */
					$$tmpl_initiator_html = apply_filters("simple_history/header_initiator_html_existing_user", $tmpl_initiator_html);

					$initiator_html .= sprintf(
						$tmpl_initiator_html,
						esc_html( $user->user_login ), // 1
						esc_html( $user->user_email ), // 2
						esc_html( $user_display_name ), // 3
						$user_role, // 4
						_x("You", "header output when initiator is the currently logged in user", "simple-history") // 5
					);

				} else if ($user_id > 0) {

					// Sender was a user, but user is deleted now
					// output all info we have
					// _user_id
					// _username
					// _user_login
					// _user_email
					$initiator_html .= sprintf(
						'<strong class="SimpleHistoryLogitem__inlineDivided">' .
							__('Deleted user (had id %1$s, email %2$s, login %3$s)', "simple-history") .
							'</strong>',
						esc_html( $context["_user_id"] ),
						esc_html( $context["_user_email"] ),
						esc_html( $context["_user_login"] )
					);

				}

				break;

			case "web_user":

				if ( empty( $context["_server_remote_addr"] ) ) {

					$initiator_html .= "<strong class='SimpleHistoryLogitem__inlineDivided'>" . __("Anonymous web user", "simple-history") . "</strong> ";

				} else {

					$iplookup_link = sprintf('https://ipinfo.io/%1$s', esc_attr( $context["_server_remote_addr"] ));

					$initiator_html .= "<strong class='SimpleHistoryLogitem__inlineDivided SimpleHistoryLogitem__anonUserWithIp'>";
					$initiator_html .= sprintf(
											__('Anonymous user from %1$s', "simple-history"),
											"<a target='_blank' href={$iplookup_link} class='SimpleHistoryLogitem__anonUserWithIp__theIp'>" . esc_attr( $context["_server_remote_addr"] ) . "</a>"
										);
					$initiator_html .= "</strong> ";

					// $initiator_html .= "<strong>" . __("<br><br>Unknown user from {$context["_server_remote_addr"]}") . "</strong>";
					// $initiator_html .= "<strong>" . __("<br><br>{$context["_server_remote_addr"]}") . "</strong>";
					// $initiator_html .= "<strong>" . __("<br><br>User from IP {$context["_server_remote_addr"]}") . "</strong>";
					// $initiator_html .= "<strong>" . __("<br><br>Non-logged in user from IP  {$context["_server_remote_addr"]}") . "</strong>";

				}

				break;

			case "other":
				$initiator_html .= "<strong class='SimpleHistoryLogitem__inlineDivided'>Other</strong>";
				break;

			// no initiator
			case null:
				// $initiator_html .= "<strong class='SimpleHistoryLogitem__inlineDivided'>Null</strong>";
				break;

			default:
				$initiator_html .= "<strong class='SimpleHistoryLogitem__inlineDivided'>" . esc_html( $initiator ) . "</strong>";

		}

		/**
	     * Filter generated html for the initiator row header html
	     *
	     * @since 2.0
	     *
	     * @param string $initiator_html
	     * @param object $row Log row
	     */
		$initiator_html = apply_filters("simple_history/row_header_initiator_output", $initiator_html, $row);


		// HTML for date
		// Date (should...) always exist
		// http://developers.whatwg.org/text-level-semantics.html#the-time-element
		$date_html = "";
		$str_when = "";
		$date_datetime = new DateTime($row->date);

		/**
	     * Filter how many seconds as most that can pass since an
	     * event occured to show "nn minutes ago" (human diff time-format) instead of exact date
	     *
	     * @since 2.0
	     *
	     * @param int $time_ago_max_time Seconds
	     */
		$time_ago_max_time = DAY_IN_SECONDS * 2;
		$time_ago_max_time = apply_filters("simple_history/header_time_ago_max_time", $time_ago_max_time);

		/**
	     * Filter how many seconds as most that can pass since an
	     * event occured to show "just now" instead of exact date
	     *
	     * @since 2.0
	     *
	     * @param int $time_ago_max_time Seconds
	     */
		$time_ago_just_now_max_time = 30;
		$time_ago_just_now_max_time = apply_filters("simple_history/header_just_now_max_time", $time_ago_just_now_max_time);

		if ( time() - $date_datetime->getTimestamp() <= $time_ago_just_now_max_time ) {

			// show "just now" if event is very recent
			$str_when = __("Just now", "simple-history");

		} else if ( time() - $date_datetime->getTimestamp() > $time_ago_max_time ) {

			/* translators: Date format for log row header, see http://php.net/date */
			$datef = __( 'M j, Y \a\t G:i', "simple-history" );
			$str_when = date_i18n( $datef, $date_datetime->getTimestamp() );

		} else {

			// Show "nn minutes ago" when event is xx seconds ago or earlier
			$date_human_time_diff = human_time_diff( $date_datetime->getTimestamp(), time() );
			/* translators: 1: last modified date and time in human time diff-format */
			$str_when = sprintf( __( '%1$s ago', 'simple-history' ), $date_human_time_diff );

		}

		$item_permalink = admin_url("index.php?page=simple_history_page");
		$item_permalink .= "#item/{$row->id}";

		$date_html = "<span class='SimpleHistoryLogitem__permalink SimpleHistoryLogitem__when SimpleHistoryLogitem__inlineDivided'>";
		$date_html .= "<a class='' href='{$item_permalink}'>";
		$date_html .= sprintf(
			'<time datetime="%1$s" title="%1$s" class="">%2$s</time>',
			$date_datetime->format(DateTime::RFC3339), // 1 datetime attribute
			$str_when
		);
		$date_html .= "</a>";
		$date_html .= "</span>";

		// Loglevel
		// SimpleHistoryLogitem--loglevel-warning
		/*
		$level_html = sprintf(
			'<span class="SimpleHistoryLogitem--logleveltag SimpleHistoryLogitem--logleveltag-%1$s">%1$s</span>',
			$row->level
		);
		*/

		// Glue together final result
		$template = '%1$s%2$s';
		#if ( ! $initiator_html ) {
		#	$template = '%2$s';
		#}

		$html = sprintf(
			$template,
			$initiator_html, // 1
			$date_html // 2
			// $level_html // 3
		);

		/**
	     * Filter generated html for the log row header
	     *
	     * @since 2.0
	     *
	     * @param string $html
	     * @param object $row Log row
	     */
		$html = apply_filters("simple_history/row_header_output", $html, $row);

		return $html;

	}

	/**
	 * Returns the plain text version of this entry
	 * Used in for example CSV-exports.
	 * Defaults to log message with context interpolated.
	 * Keep format as plain and simple as possible.
	 * Links are ok, for example to link to users or posts.
	 * Tags will be stripped when text is used for CSV-exports and so on.
	 * Keep it on a single line. No <p> or <br> and so on.
	 *
	 * Example output:
	 * Edited post "About the company"
	 *
	 * Message should sound like it's coming from the user.
	 * Image that the name of the user is added in front of the text:
	 * Jessie James: Edited post "About the company"
	 */
	public function getLogRowPlainTextOutput($row) {

		$message = $row->message;
		$message_key = $row->context["_message_key"];

		// Message is translated here, but translation must be added in
		// plain text before

		if ( empty( $message_key ) ) {

			// Leave message alone

		} else {

			$message = $this->messages[ $message_key ]["translated_text"];

		}

		$html = $this->interpolate( $message, $row->context );

		// All messages are escaped by default.
		// If you need unescaped output override this method
		// in your own logger
		$html = esc_html($html);

		/**
	     * Filter generated output for plain text output
	     *
	     * @since 2.0
	     *
	     * @param string $html
	     * @param object $row Log row
	     */
		$html = apply_filters("simple_history/row_plain_text_output", $html, $row);

		return $html;

	}

	/**
	 * Get output for image
	 * Image can be for example gravar if sender is user,
	 * or other images if sender i system, wordpress, and so on
	 */
	public function getLogRowSenderImageOutput($row) {

		$sender_image_html = "";
		$sender_image_size = 32;

		$initiator = $row->initiator;

		switch ( $initiator ) {

			// wp_user = wordpress uses, but user may have been deleted since log entry was added
			case "wp_user":

				$user_id = isset($row->context["_user_id"]) ? $row->context["_user_id"] : null;

				if ( $user_id > 0 && $user = get_user_by("id", $user_id) ) {

					// Sender was user
					$sender_image_html = $this->simpleHistory->get_avatar( $user->user_email, $sender_image_size );

				} else if ($user_id > 0) {

					// Sender was a user, but user is deleted now
					$sender_image_html = $this->simpleHistory->get_avatar( "", $sender_image_size );

				} else {

					$sender_image_html = $this->simpleHistory->get_avatar( "", $sender_image_size );

				}

				break;

		}
		/**
	     * Filter generated output for row image (sender image)
	     *
	     * @since 2.0
	     *
	     * @param string $sender_image_html
	     * @param object $row Log row
	     */
		$sender_image_html = apply_filters("simple_history/row_sender_image_output", $sender_image_html, $row);

		return $sender_image_html;

	}

	/**
	 * Use this method to output detailed output for a log row
	 * Example usage is if a user has uploaded an image then a
	 * thumbnail of that image can bo outputed here
	 *
	 * @param object $row
	 * @return string HTML-formatted output
	 */
	public function getLogRowDetailsOutput($row) {

		$html = "";

		/**
	     * Filter generated output for details
	     *
	     * @since 2.0
	     *
	     * @param string $html
	     * @param object $row Log row
	     */
		$html = apply_filters("simple_history/row_details_output", $html, $row);

		return $html;

	}


	/**
	 * System is unusable.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public static function emergency($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::EMERGENCY, $message, $context);

	}

	/**
	 * System is unusable.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function emergencyMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::EMERGENCY, $message, $context);

	}


	/**
	 * Action must be taken immediately.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public static function alert($message, array $context = array())
	{
		return $this->log(SimpleLoggerLogLevels::ALERT, $message, $context);

	}

	/**
	 * Action must be taken immediately.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function alertMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::ALERT, $message, $context);

	}


	/**
	 * Critical conditions.
	 *
	 * Example: Application component unavailable, unexpected exception.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public static function critical($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::CRITICAL, $message, $context);

	}

	/**
	 * Critical conditions.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function criticalMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::CRITICAL, $message, $context);

	}


	/**
	 * Runtime errors that do not require immediate action but should typically
	 * be logged and monitored.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function error($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::ERROR, $message, $context);

	}

	/**
	 * Runtime errors that do not require immediate action but should typically
	 * be logged and monitored.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function errorMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::ERROR, $message, $context);

	}


	/**
	 * Exceptional occurrences that are not errors.
	 *
	 * Example: Use of deprecated APIs, poor use of an API, undesirable things
	 * that are not necessarily wrong.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function warning($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::WARNING, $message, $context);

	}

	/**
	 * Exceptional occurrences that are not errors.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function warningMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::WARNING, $message, $context);

	}


	/**
	 * Normal but significant events.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function notice($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::NOTICE, $message, $context);

	}

	/**
	 * Normal but significant events.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function noticeMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::NOTICE, $message, $context);

	}


	/**
	 * Interesting events.
	 *
	 * Example: User logs in, SQL logs.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function info($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::INFO, $message, $context);

	}

	/**
	 * Interesting events.
	 *
	 * Example: User logs in, SQL logs.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function infoMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::INFO, $message, $context);

	}

	/**
	 * Detailed debug information.
	 *
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function debug($message, array $context = array())
	{

		return $this->log(SimpleLoggerLogLevels::DEBUG, $message, $context);

	}

	/**
	 * Detailed debug information.
	 *
	 * @param string $message key from getInfo messages array
	 * @param array $context
	 * @return null
	 */
	public function debugMessage($message, array $context = array())
	{

		if ( ! isset( $this->messages[ $message ]["untranslated_text"] ) ) {
			return;
		}

		$context["_message_key"] = $message;
		$message = $this->messages[ $message ]["untranslated_text"];

		$this->log(SimpleLoggerLogLevels::DEBUG, $message, $context);

	}

	/**
	 * Logs with an arbitrary level.
	 *
	 * @param mixed $level
	 * @param string $message
	 * @param array $context
	 * @return null
	 */
	public function log($level, $message, array $context = array())
	{

		global $wpdb;

		/**
	     * Filter arguments passed to log funtion
	     *
	     * @since 2.0
	     *
	     * @param string $level
	     * @param string $message
	     * @param array $context
	     */
		apply_filters("simple_history/log_arguments", $level, $message, $context);

		/* Store date at utc or local time?
		 * Some info here:
		 * http://www.skyverge.com/blog/down-the-rabbit-hole-wordpress-and-timezones/
		 * UNIX timestamp = no timezone = UTC
		 * anything is better than now() anyway!
		 * WP seems to use the local time, so I will go with that too I think
		 * GMT/UTC-time is: date_i18n($timezone_format, false, 'gmt'));
		 * local time is: date_i18n($timezone_format));
		 */
		$localtime = current_time("mysql", 1);

		$db_table = $wpdb->prefix . SimpleHistory::DBTABLE;

		/**
	     * Filter db table used for simple history events
	     *
	     * @since 2.0
	     *
	     * @param string $db_table
	     */
		$db_table = apply_filters("simple_history/db_table", $db_table);

		$data = array(
			"logger" => $this->slug,
			"level" => $level,
			"date" => $localtime,
			"message" => $message,
		);

		// Allow date to be override
		// Date must be in format 'Y-m-d H:i:s'
		if ( isset( $context["_date"] ) ) {
			$data["date"] = $context["_date"];
			unset($context["_date"]);
		}

		// Add occasions id
		$occasions_id = null;
		if ( isset( $context["_occasionsID"] ) ) {

			// Minimize risk of similar loggers logging same messages and such and resulting in same occasions id
			// by always adding logger slug
			$occasions_data = array(
				"_occasionsID" => $context["_occasionsID"],
				"_loggerSlug" => $this->slug
			);
			$occasions_id = md5( json_encode($occasions_data) );
			unset( $context["_occasionsID"] );

		} else {

			// No occasions id specified, create one bases on the data array
			$occasions_data = $data + $context;

			// Don't include date in context data
			unset($occasions_data["date"]);

			//sf_d($occasions_data);exit;
			$occasions_id = md5( json_encode($occasions_data) );

		}

		$data["occasionsID"] = $occasions_id;

		// Log event type, defaults to other if not set
		/*
		if ( isset( $context["_type"] ) ) {
			$data["type"] = $context["_type"];
			unset( $context["_type"] );
		} else {
			$data["type"] = SimpleLoggerLogTypes::OTHER;
		}
		*/

		// Log initiator, defaults to current user if exists, or other if not user exist
		if ( isset( $context["_initiator"] ) ) {

			// Manually set in context
			$data["initiator"] = $context["_initiator"];
			unset( $context["_initiator"] );

		} else {

			// No initiator set.

			$data["initiator"] = SimpleLoggerLogInitiators::OTHER;

			// Check if user is responsible.
			if ( function_exists("wp_get_current_user") ) {

				$current_user = wp_get_current_user();

				if ( isset( $current_user->ID ) && $current_user->ID) {

					$data["initiator"] = SimpleLoggerLogInitiators::WP_USER;;
					$context["_user_id"] = $current_user->ID;
					$context["_user_login"] = $current_user->user_login;
					$context["_user_email"] = $current_user->user_email;

				}

			}

			// If cron then set WordPress as responsible
			if ( defined('DOING_CRON') && DOING_CRON ) {

				// Seems to be wp cron running and doing this
				$data["initiator"] = SimpleLoggerLogInitiators::WORDPRESS;
				$context["_wordpress_cron_running"] = true;

			}

		}

		/**
	     * Filter data to be saved to db
	     *
	     * @since 2.0
	     *
	     * @param array $data
	     */
		$data = apply_filters("simple_history/log_insert_data", $data);

		// Insert data into db
		// sf_d($db_table, '$db_table');exit;
		$result = $wpdb->insert( $db_table, $data );

		// Only save context if able to store row
		if ( false === $result ) {

			$history_inserted_id = null;

		} else {

			$history_inserted_id = $wpdb->insert_id;

			$db_table_contexts = $wpdb->prefix . SimpleHistory::DBTABLE_CONTEXTS;

			/**
		     * Filter table name for contexts
		     *
		     * @since 2.0
		     *
		     * @param string $db_table_contexts
		     */
			$db_table_contexts = apply_filters("simple_history/logger_db_table_contexts", $db_table_contexts);

			if ( ! is_array($context) ) {
				$context = array();
			}

			// Append user id to context, if not already added
			if ( ! isset( $context["_user_id"] ) ) {

				// wp_get_current_user is ont available early
				// http://codex.wordpress.org/Function_Reference/wp_get_current_user
				// https://core.trac.wordpress.org/ticket/14024
				if ( function_exists("wp_get_current_user") ) {

					$current_user = wp_get_current_user();

					if ( isset( $current_user->ID ) && $current_user->ID) {
						$context["_user_id"] = $current_user->ID;
						$context["_user_login"] = $current_user->user_login;
						$context["_user_email"] = $current_user->user_email;
					}

				}

			}

			// Add remote addr to context
			// Good to always have
			if ( ! isset( $context["_server_remote_addr"] ) ) {

				// If web server is behind a load balancer then the ip address will always be the same
				// apache_request_headers
				// See bug report: https://wordpress.org/support/topic/use-x-forwarded-for-http-header-when-logging-remote_addr?replies=1#post-6422981
				if ( function_exists("getallheaders") ) {
					
					$all_headers = getallheaders();
					if ( ! empty( $all_headers["X-Forwarded-For"] ) ) {
						$context["_server_remote_addr"] = $all_headers["X-Forwarded-For"];
					}

				}

				// Still no remote addr? Then simply add remote addr
				if ( empty( $context["_server_remote_addr"] ) ) {
					$context["_server_remote_addr"] = $_SERVER["REMOTE_ADDR"];
				}

			}

			// Append http referer
			// Also good to always have!
			if ( ! isset( $context["_server_http_referer"] ) && isset( $_SERVER["HTTP_REFERER"] ) ) {
				$context["_server_http_referer"] = $_SERVER["HTTP_REFERER"];
			}

			// Insert all context values into db
			foreach ( $context as $key => $value ) {

				$data = array(
					"history_id" => $history_inserted_id,
					"key" => $key,
					"value" => $value,
				);

				$result = $wpdb->insert( $db_table_contexts, $data );

			}

		}

		$this->lastInsertID = $history_inserted_id;

		// Return $this so we can chain methods
		return $this;

	} // log

	/**
	 * Override this to add CSS in <head> for your logger.
	 * The CSS that you output will only be outputed
	 * on pages where Simple History is used.
	 */
	function adminCSS() {
		/*
		?>
		<style>
			body {
				border: 2px solid red;
			}
		</style>
		<?php
		*/
	}

	/**
	 * Override this to add JavaScript in the footer for your logger.
	 * The JS that you output will only be outputed
	 * on pages where Simple History is used.
	 */
	function adminJS() {
		/*
		?>
		<script>
			console.log("This is outputed in the footer");
		</script>
		<?php
		*/
	}

}

/**
 * Describes log initiator, i.e. who caused to log event to happend
 */
class SimpleLoggerLogInitiators
{

	// A wordpress user that at the log event created did exist in the wp database
	// May have been deleted when the log is viewed
	const WP_USER = 'wp_user';

	// Cron job run = wordpress initiated
	// Email sent to customer on webshop = system/wordpress/anonymous web user
	// Javascript error occured on website = anonymous web user
	const WEB_USER = 'web_user';

	// WordPress core or plugins updated automatically via wp-cron
	const WORDPRESS = "wp";

	// I dunno
	const OTHER = 'other';
}


/**
 * Describes log event type
 * Based on the CRUD-types
 * http://en.wikipedia.org/wiki/Create,_read,_update_and_delete
 * More may be added later on if needed
 * Note: not in use at the moment
 */
class SimpleLoggerLogTypes
{
	const CREATE = 'create';
	const READ = 'read';
	const UPDATE = 'update';
	const DELETE = 'delete';
	const OTHER = 'other';
}

/**
 * Describes log levels
 */
class SimpleLoggerLogLevels
{
	const EMERGENCY = 'emergency';
	const ALERT = 'alert';
	const CRITICAL = 'critical';
	const ERROR = 'error';
	const WARNING = 'warning';
	const NOTICE = 'notice';
	const INFO = 'info';
	const DEBUG = 'debug';
}
